@@ -50,10 +50,15 @@
 static struct gvm_configuration config;
 static LibHalContext *hal_ctx;
 
-/** Table of UDI's for volumes being mounted by g-v-m that we need to apply policy to */
+struct _MountPolicy {
+	char *udi;
+	int apply;
+};
+
+/* Table of UDI's for volumes being mounted by g-v-m that we need to apply policy to */
 static GHashTable *mount_table = NULL;
 
-/** List of UDI's of all volumes mounted during the lifetime of the program */
+/* List of UDI's of all volumes mounted during the lifetime of the program */
 static GSList *mounted_volumes = NULL;
 
 
@@ -73,12 +78,16 @@ enum {
 	AUTOMOUNT_MEDIA,
 	AUTOPHOTO,
 	AUTOPHOTO_COMMAND,
+	AUTOPILOT,
+	AUTOPILOT_COMMAND,
 	AUTOPLAY_CDA,
 	AUTOPLAY_CDA_COMMAND,
 	AUTOPLAY_DVD,
 	AUTOPLAY_DVD_COMMAND,
 	AUTOPLAY_VCD,
 	AUTOPLAY_VCD_COMMAND,
+	AUTOPOCKETPC,
+	AUTOPOCKETPC_COMMAND,
 	AUTOPRINTER,
 	AUTOPRINTER_COMMAND,
 	AUTORUN,
@@ -101,12 +110,16 @@ static struct {
 	{ GCONF_ROOT "automount_media",           TYPE_BOOL,   &config.automount_media           },
 	{ GCONF_ROOT "autophoto",                 TYPE_BOOL,   &config.autophoto                 },
 	{ GCONF_ROOT "autophoto_command",         TYPE_STRING, &config.autophoto_command         },
+	{ GCONF_ROOT "autopalmsync",              TYPE_BOOL,   &config.autopilot                 },
+	{ GCONF_ROOT "autopalmsync_command",      TYPE_STRING, &config.autopilot_command         },
 	{ GCONF_ROOT "autoplay_cda",              TYPE_BOOL,   &config.autoplay_cda              },
 	{ GCONF_ROOT "autoplay_cda_command",      TYPE_STRING, &config.autoplay_cda_command      },
 	{ GCONF_ROOT "autoplay_dvd",              TYPE_BOOL,   &config.autoplay_dvd              },
 	{ GCONF_ROOT "autoplay_dvd_command",      TYPE_STRING, &config.autoplay_dvd_command      },
 	{ GCONF_ROOT "autoplay_vcd",              TYPE_BOOL,   &config.autoplay_vcd              },
 	{ GCONF_ROOT "autoplay_vcd_command",      TYPE_STRING, &config.autoplay_vcd_command      },
+	{ GCONF_ROOT "autopocketpc",              TYPE_BOOL,   &config.autopocketpc              },
+	{ GCONF_ROOT "autopocketpc_command",      TYPE_STRING, &config.autopocketpc_command      },
 	{ GCONF_ROOT "autoprinter",               TYPE_BOOL,   &config.autoprinter               },
 	{ GCONF_ROOT "autoprinter_command",       TYPE_STRING, &config.autoprinter_command       },
 	{ GCONF_ROOT "autorun",                   TYPE_BOOL,   &config.autorun                   },
@@ -184,6 +197,20 @@ typedef enum {
 	GVM_PROMPT_BURN_DVD,
 } GvmPrompt;
 
+typedef struct _GvmPromptCtx GvmPromptCtx;
+typedef void (* GvmPromptCallback) (GvmPromptCtx *ctx, int action);
+
+struct _GvmPromptCtx {
+	GvmPrompt prompt;
+	
+	GvmPromptCallback callback;
+	
+	char *udi;
+	char *device;
+	char *mount_point;
+	char *path;
+};
+
 static struct {
 	GtkDialogFlags flags;
 	
@@ -251,32 +278,71 @@ static struct {
 	  GVM_BUTTONS_BURN_CD, G_N_ELEMENTS (GVM_BUTTONS_BURN_CD),
 	  GVM_RESPONSE_BURN_DATA_CD,
 	  N_("Choose Disc Type"),
-	  N_("You have inserted an empty disc."),
+	  N_("You have inserted a blank disc."),
 	  N_("What would you like to do?"), FALSE, FALSE,
 	  NULL, NULL },
 	{ 0, "gnome-dev-disc-dvdr", NULL,
 	  GVM_BUTTONS_BURN_DVD, G_N_ELEMENTS (GVM_BUTTONS_BURN_DVD),
 	  GVM_RESPONSE_BURN_DATA_CD,
 	  N_("Choose Disc Type"),
-	  N_("You have inserted an empty disc."),
+	  N_("You have inserted a blank disc."),
 	  N_("What would you like to do?"), FALSE, FALSE,
 	  NULL, NULL },
 };
 
-static void
-prompt_response_cb (GtkWidget *dialog, int response, gpointer user_data)
+
+static GvmPromptCtx *
+gvm_prompt_ctx_new (GvmPrompt prompt, GvmPromptCallback callback, const char *udi,
+		    const char *device, const char *mount_point, const char *path)
 {
-	int prompt = GPOINTER_TO_INT (user_data);
+	GvmPromptCtx *ctx;
+	
+	ctx = g_malloc (sizeof (GvmPromptCtx));
+	ctx->prompt = prompt;
+	ctx->callback = callback;
+	ctx->udi = g_strdup (udi);
+	ctx->device = g_strdup (device);
+	ctx->mount_point = g_strdup (mount_point);
+	ctx->path = g_strdup (path);
+	
+	return ctx;
+}
+
+static void
+gvm_prompt_ctx_free (GvmPromptCtx *ctx)
+{
+	g_free (ctx->udi);
+	g_free (ctx->device);
+	g_free (ctx->mount_point);
+	g_free (ctx->path);
+	g_free (ctx);
+}
+
+static void
+prompt_response_cb (GtkWidget *dialog, int response, GvmPromptCtx *ctx)
+{
+	GtkToggleButton *checkbox;
+	GConfClient *gconf;
 	GError *err = NULL;
 	
-	if (response != GTK_RESPONSE_HELP)
-		return;
-	
-	g_signal_stop_emission_by_name (dialog, "response");
-	gnome_url_show (gvm_prompts[prompt].help_uri, &err);
-	if (err) {
-		warn ("Unable to run help uri: %s", err->message);
-		g_error_free (err);
+	if (response == GTK_RESPONSE_HELP) {
+		g_signal_stop_emission_by_name (dialog, "response");
+		gnome_url_show (gvm_prompts[ctx->prompt].help_uri, &err);
+		if (err) {
+			warn ("Unable to run help uri: %s", err->message);
+			g_error_free (err);
+		}
+	} else {
+		checkbox = g_object_get_data ((GObject *) dialog, "checkbox");
+		if (checkbox && gtk_toggle_button_get_active (checkbox) && response != GTK_RESPONSE_CANCEL) {
+			gconf = gconf_client_get_default ();
+			gconf_client_set_int (gconf, gvm_prompts[ctx->prompt].ask_again_key, response, NULL);
+			g_object_unref (gconf);
+		}
+		
+		ctx->callback (ctx, response);
+		gtk_widget_destroy (dialog);
+		gvm_prompt_ctx_free (ctx);
 	}
 }
 
@@ -320,10 +386,11 @@ argv_expand (const char *format, int argc, char **argv)
 	return str;
 }
 
-static int
-gvm_prompt (GvmPrompt prompt, int argc, char **argv)
+static void
+gvm_prompt (GvmPromptCtx *ctx, int argc, char **argv)
 {
 	GtkWidget *dialog, *hbox, *vbox, *image, *label, *check = NULL;
+	GvmPrompt prompt = ctx->prompt;
 	GnomeIconTheme *theme;
 	GConfClient *gconf;
 	GError *err = NULL;
@@ -336,8 +403,11 @@ gvm_prompt (GvmPrompt prompt, int argc, char **argv)
 	/* don't prompt the user again if she's already chosen a default action and has asked to not be prompetd again */
 	if (gvm_prompts[prompt].ask_again_key) {
 		response = gconf_client_get_int (gconf, gvm_prompts[prompt].ask_again_key, &err);
-		if (response > GVM_RESPONSE_NONE && err == NULL)
-			return response;
+		if (response > GVM_RESPONSE_NONE && err == NULL) {
+			ctx->callback (ctx, response);
+			return;
+		}
+		
 		if (err != NULL)
 			g_error_free (err);
 	}
@@ -355,10 +425,8 @@ gvm_prompt (GvmPrompt prompt, int argc, char **argv)
 	if (gvm_prompts[prompt].flags & GTK_DIALOG_MODAL)
 		gtk_window_set_modal ((GtkWindow *) dialog, TRUE);
 	
-	if (gvm_prompts[prompt].help_uri) {
+	if (gvm_prompts[prompt].help_uri)
 		gtk_dialog_add_button ((GtkDialog *) dialog, GTK_STOCK_HELP, GTK_RESPONSE_HELP);
-		g_signal_connect (dialog, "response", G_CALLBACK (prompt_response_cb), GINT_TO_POINTER (prompt));
-	}
 	
 	if (gvm_prompts[prompt].buttons) {
 		for (i = 0; i < gvm_prompts[prompt].n_buttons; i++) {
@@ -437,22 +505,18 @@ gvm_prompt (GvmPrompt prompt, int argc, char **argv)
 		check = gtk_check_button_new_with_mnemonic (_(gvm_prompts[prompt].ask_again_label));
 		gtk_container_set_border_width ((GtkContainer *) check, 0);
 		gtk_box_pack_start ((GtkBox *) vbox, check, FALSE, FALSE, 0);
+		g_object_set_data ((GObject *) dialog, "checkbox", check);
 		gtk_widget_show (check);
 	}
 	
 	gtk_widget_show (vbox);
 	gtk_widget_show (hbox);
 	
-	response = gtk_dialog_run ((GtkDialog *) dialog);
-	
-	if (check != NULL && gtk_toggle_button_get_active ((GtkToggleButton *) check)
-	    && response != GTK_RESPONSE_CANCEL)
-		gconf_client_set_int (gconf, gvm_prompts[prompt].ask_again_key, response, NULL);
-	
-	gtk_widget_destroy (dialog);
 	g_object_unref (gconf);
 	
-	return response;
+	g_signal_connect (dialog, "response", G_CALLBACK (prompt_response_cb), ctx);
+	
+	gtk_widget_show (dialog);
 }
 
 static void
@@ -462,11 +526,19 @@ to_be_or_not_to_be (void)
 	 * If all of the options that control our policy are disabled, then we
 	 * have no point in living.  Save the user some memory and exit.
 	 */
-	if (!(config.automount_drives || config.autobrowse || config.autorun 
-	      || config.autoplay_cda || config.autoplay_dvd 
-	      || config.autophoto || config.autoipod
-	      || config.autoprinter)) {
-		dbg ("daemon exit: no point living\n");
+	/* you used to say live and let live... */
+	gboolean live = FALSE;
+	size_t i;
+	
+	/* ...but in this ever changing world in which we live in... */
+	for (i = 0; i < G_N_ELEMENTS (gvm_settings) && !live; i++) {
+		if (gvm_settings[i].type == TYPE_BOOL)
+			live = *((int *) gvm_settings[i].var);
+	}
+	
+	/* makes you give it a cry... */
+	if (!live) {
+		dbg ("daemon exit: live and let die\n");
 		exit (EXIT_SUCCESS);
 	}
 }
@@ -639,24 +711,6 @@ gvm_run_command (const char *command, const char *udi, const char *device, const
 	return TRUE;
 }
 
-/*
- * gvm_ask_autorun - ask the user if they want to autorun a specific file
- *
- * Returns TRUE if the user selected 'Run' and FALSE otherwise
- */
-static gboolean
-gvm_ask_autorun (const char *path)
-{
-	gboolean autorun;
-	char *argv[1];
-	
-	argv[0] = g_path_get_basename (path);
-	autorun = gvm_prompt (GVM_PROMPT_AUTORUN, 1, argv) == GVM_RESPONSE_RUN;
-	g_free (argv[0]);
-	
-	return autorun;
-}
-
 static gboolean
 gvm_check_dir (const char *dirname, const char *name)
 {
@@ -671,7 +725,7 @@ gvm_check_dir (const char *dirname, const char *name)
 	
 	while ((dent = readdir (dir))) {
 		if (!g_ascii_strcasecmp (dent->d_name, name)) {
-			path = g_build_filename (dirname, dent->d_name);
+			path = g_build_filename (dirname, dent->d_name, NULL);
 			if (stat (path, &st) == 0 && S_ISDIR (st.st_mode))
 				exists = TRUE;
 			g_free (path);
@@ -769,8 +823,11 @@ gvm_udi_is_ptp_camera (const char *udi)
 	gboolean is_ptp_camera = FALSE;
 	char *access_method;
 	
+#if 0
+	/* Note: this query is not necessary atm because our only caller already has this info */
 	if (!libhal_device_query_capability (hal_ctx, udi, "camera", NULL))
 		return FALSE;
+#endif
 	
 	if (!(access_method = libhal_device_get_property_string (hal_ctx, udi, "camera.access_method", NULL)))
 		return FALSE;
@@ -870,6 +927,57 @@ gvm_run_ipod (const char *udi, const char *device, const char *mount_point)
 		gvm_run_command (config.autoipod_command, udi, device, mount_point);
 }
 
+
+/*
+ * gvm_run_pilot - sync the pda palm pilot
+ */
+static void
+gvm_run_pilot (const char *udi)
+{
+	DBusError error;
+	char *device;
+	
+	if (config.autopilot_command == NULL)
+		return;
+	
+	dbus_error_init (&error);
+	/* pda.palm.hotsync_interface should be the same as serial.device, but just in case... */
+	if (!(device = libhal_device_get_property_string (hal_ctx, udi, "pda.palm.hotsync_interface", &error))) {
+		warn ("cannot get pda.palm.hotsync_interface property: %s", error.message);
+		if (dbus_error_is_set (&error))
+			dbus_error_free (&error);
+		return;
+	}
+	
+	gvm_run_command (config.autopilot_command, udi, device, NULL);
+	libhal_free_string (device);
+}
+
+/*
+ * gvm_run_pilot - sync the pda palm pilot
+ */
+static void
+gvm_run_pocketpc (const char *udi)
+{
+	DBusError error;
+	char *device;
+	
+	if (config.autopocketpc_command == NULL)
+		return;
+	
+	dbus_error_init (&error);
+	/* pda.palm.hotsync_interface should be the same as serial.device, but just in case... */
+	if (!(device = libhal_device_get_property_string (hal_ctx, udi, "pda.pocketpc.hotsync_interface", &error))) {
+		warn ("cannot get pda.pocketpc.hotsync_interface property: %s", error.message);
+		if (dbus_error_is_set (&error))
+			dbus_error_free (&error);
+		return;
+	}
+	
+	gvm_run_command (config.autopocketpc_command, udi, device, NULL);
+	libhal_free_string (device);
+}
+
 /*
  * gvm_run_printer - launch the printer application
  */
@@ -894,6 +1002,36 @@ gvm_run_printer (const char *udi)
 	libhal_free_string (device);
 }
 
+static void
+import_photos_cb (GvmPromptCtx *ctx, int action)
+{
+	if (action == GVM_RESPONSE_IMPORT_PHOTOS)
+		gvm_run_camera (ctx->udi, ctx->device, ctx->mount_point);
+}
+
+static void
+autorun_cb (GvmPromptCtx *ctx, int action)
+{
+	gboolean autobrowse = TRUE;
+	GError *error = NULL;
+	char *argv[2];
+	
+	if (action == GVM_RESPONSE_RUN) {
+		argv[0] = ctx->path;
+		argv[1] = NULL;
+		
+		g_spawn_async (g_get_home_dir (), argv, NULL,
+			       0, NULL, NULL, NULL, &error);
+		if (error)
+			warn ("failed to exec %s: %s", ctx->path, error->message);
+		else
+			autobrowse = FALSE;
+	}
+	
+	if (config.autobrowse && autobrowse)
+		gvm_run_command (NAUTILUS_COMMAND, ctx->udi, ctx->device, ctx->mount_point);
+}
+
 /*
  * gvm_autorun - automatically execute stuff
  *
@@ -902,7 +1040,8 @@ gvm_run_printer (const char *udi)
 static void
 gvm_autorun (const char *udi, const char *device, const char *mount_point)
 {
-	gboolean autorun_succeeded = FALSE;
+	gboolean autobrowse = FALSE;
+	GvmPromptCtx *ctx;
 	
 	if (gvm_check_dvd (udi, device, mount_point))
 		return;
@@ -911,51 +1050,39 @@ gvm_autorun (const char *udi, const char *device, const char *mount_point)
 		return;
 	
 	if (config.autophoto && gvm_check_photos (udi, device, mount_point)) {
-		if (gvm_prompt (GVM_PROMPT_IMPORT_PHOTOS, 0, NULL) == GVM_RESPONSE_IMPORT_PHOTOS) {
-			gvm_run_camera (udi, device, mount_point);
-			return;
-		}
+		ctx = gvm_prompt_ctx_new (GVM_PROMPT_IMPORT_PHOTOS, import_photos_cb, udi, device, mount_point, NULL);
+		gvm_prompt (ctx, 0, NULL);
+		return;
 	}
 	
 	if (config.autorun && config.autorun_path) {
-		char **autorun_fns;
+		char **autorun;
 		int i;
 		
-		autorun_fns = g_strsplit (config.autorun_path, ":", -1);
+		autorun = g_strsplit (config.autorun_path, ":", -1);
 		
-		for (i = 0; autorun_fns[i]; i++) {
-			char *path, *argv[2];
-			
-			path = g_strdup_printf ("%s/%s", mount_point, autorun_fns[i]);
-			argv[0] = path;
-			argv[1] = NULL;
+		for (i = 0; autorun[i]; i++) {
+			char *argv[1];
+			char *path;
 			
+			path = g_strdup_printf ("%s/%s", mount_point, autorun[i]);
 			if (access (path, X_OK)) {
 				g_free (path);
 				continue;
 			}
 			
-			if (gvm_ask_autorun (path)) {
-				GError *error = NULL;
-				
-				g_spawn_async (g_get_home_dir (), argv, NULL,
-					       0, NULL, NULL, NULL, &error);
-				if (error)
-					warn ("failed to exec %s: %s", path, error->message);
-				else
-					autorun_succeeded = TRUE;
-				
-				g_free (path);
-				break;
-			}
-			
+			argv[0] = autorun[i];
+			ctx = gvm_prompt_ctx_new (GVM_PROMPT_AUTORUN, autorun_cb, udi, device, mount_point, path);
+			gvm_prompt (ctx, 1, argv);
+			autobrowse = FALSE;
 			g_free (path);
+			break;
 		}
 		
-		g_strfreev (autorun_fns);
+		g_strfreev (autorun);
 	}
 	
-	if (config.autobrowse && !autorun_succeeded)
+	if (config.autobrowse && autobrowse)
 		gvm_run_command (NAUTILUS_COMMAND, udi, device, mount_point);
 }
 
@@ -981,6 +1108,36 @@ gvm_udi_is_subfs_mount (const char *udi)
 	return subfs;
 }
 
+static void
+import_camera_cb (GvmPromptCtx *ctx, int action)
+{
+	switch (action) {
+	case GVM_RESPONSE_IMPORT_PHOTOS:
+		gvm_run_camera (ctx->udi, ctx->device, ctx->mount_point);
+		break;
+	case GVM_RESPONSE_BROWSE:
+		gvm_run_command (NAUTILUS_COMMAND, ctx->udi, ctx->device, ctx->mount_point);
+		break;
+	default:
+		break;
+	}
+}
+
+static void
+ipod_photo_cb (GvmPromptCtx *ctx, int action)
+{
+	switch (action) {
+	case GVM_RESPONSE_IMPORT_PHOTOS:
+		gvm_run_camera (ctx->udi, ctx->device, ctx->path);
+		break;
+	case GVM_RESPONSE_SYNC_MUSIC:
+		gvm_run_ipod (ctx->udi, ctx->device, ctx->mount_point);
+		break;
+	default:
+		break;
+	}
+}
+
 /*
  * gvm_device_mounted - called once a device has been
  * mounted. Launches any user-specified applications that require the
@@ -992,6 +1149,7 @@ static void
 gvm_device_mounted (const char *udi)
 {
 	char *device = NULL, *mount_point = NULL;
+	GvmPromptCtx *ctx;
 	DBusError error;
 	
 	dbus_error_init (&error);
@@ -1011,42 +1169,26 @@ gvm_device_mounted (const char *udi)
 	
 	/* this is where the magic happens */
 	if (config.autophoto && gvm_udi_is_camera (udi)) {
-		switch (gvm_prompt (GVM_PROMPT_IMPORT_CAMERA, 0, NULL)) {
-		case GVM_RESPONSE_IMPORT_PHOTOS:
-			gvm_run_camera (udi, device, mount_point);
-			break;
-		case GVM_RESPONSE_BROWSE:
-			gvm_run_command (NAUTILUS_COMMAND, udi, device, mount_point);
-			break;
-		default:
-			break;
-		}
+		ctx = gvm_prompt_ctx_new (GVM_PROMPT_IMPORT_CAMERA, import_camera_cb, udi, device, mount_point, NULL);
+		gvm_prompt (ctx, 0, NULL);
 	} else if (config.autoipod && gvm_udi_is_ipod (udi)) {
-		int action = GVM_RESPONSE_SYNC_MUSIC;
 		char *ipod_control;
 		
 		ipod_control = g_build_filename (mount_point, "iPod_Control", NULL);
 		
 		if (gvm_check_photos (udi, device, ipod_control)) {
 			/* we have ourselves an iPod Photo - need to prompt what to do */
-			action = gvm_prompt (GVM_PROMPT_IPOD_PHOTO, 0, NULL);
-		}
-		
-		switch (action) {
-		case GVM_RESPONSE_IMPORT_PHOTOS:
-			gvm_run_camera (udi, device, ipod_control);
-			break;
-		case GVM_RESPONSE_SYNC_MUSIC:
+			ctx = gvm_prompt_ctx_new (GVM_PROMPT_IPOD_PHOTO, ipod_photo_cb, udi, device, mount_point, ipod_control);
+			gvm_prompt (ctx, 0, NULL);
+		} else {
 			gvm_run_ipod (udi, device, mount_point);
-			break;
-		default:
-			break;
 		}
 		
 		g_free (ipod_control);
 	} else {
 		gvm_autorun (udi, device, mount_point);
 	}
+	
  out:
 	
 	libhal_free_string (mount_point);
@@ -1060,20 +1202,24 @@ gvm_device_mounted (const char *udi)
  * @return TRUE iff the mount was succesful
  */
 static gboolean
-gvm_device_mount (const char *udi, const char *device, const char *mount_point)
+gvm_device_mount (const char *udi, const char *device, const char *mount_point, int apply_policy)
 {
-	char *key;
+	struct _MountPolicy *policy;
 	
-	key = g_strdup (udi);
-	g_hash_table_insert (mount_table, key, key);
+	policy = g_new (struct _MountPolicy, 1);
+	policy->udi = g_strdup (udi);
+	policy->apply = apply_policy;
+	
+	g_hash_table_insert (mount_table, policy->udi, policy);
 	
 	dbg ("mounting %s...\n", udi);
 	if (gvm_run_command (MOUNT_COMMAND, udi, device, mount_point))
 		return TRUE;
 	
 	dbg ("mount failed: %s\n", udi);
-	g_hash_table_remove (mount_table, key);
-	g_free (key);
+	g_hash_table_remove (mount_table, policy->udi);
+	g_free (policy->udi);
+	g_free (policy);
 	
 	return FALSE;
 }
@@ -1100,6 +1246,22 @@ gvm_run_cdplayer (const char *udi, const char *device, const char *mount_point)
 		gvm_run_command (config.autoplay_cda_command, udi, device, mount_point);
 }
 
+static void
+cda_extra_cb (GvmPromptCtx *ctx, int action)
+{
+	switch (action) {
+	case GVM_RESPONSE_BROWSE:
+		if (!gvm_udi_is_subfs_mount (ctx->udi))
+			gvm_device_mount (ctx->udi, ctx->device, NULL, TRUE);
+		break;
+	case GVM_RESPONSE_PLAY:
+		gvm_run_cdplayer (ctx->udi, ctx->device, NULL);
+		break;
+	default:
+		break;
+	}
+}
+
 /*
  * gvm_ask_mixed - if a mixed mode CD (CD Plus) is inserted, we can either
  * mount the data tracks or play the audio tracks.  How we handle that depends
@@ -1109,43 +1271,28 @@ gvm_run_cdplayer (const char *udi, const char *device, const char *mount_point)
 static void
 gvm_ask_mixed (const char *udi)
 {
-	char *device = NULL, *mount_point = NULL;
-	int action = GVM_RESPONSE_NONE;
+	char *device = NULL;
+	GvmPromptCtx *ctx;
 	DBusError error;
 	
 	dbus_error_init (&error);
-	device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error);
-	if (!device) {
+	if (!(device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error))) {
 		warn ("cannot get block.device: %s", error.message);
-		if (dbus_error_is_set (&error))
-			dbus_error_free (&error);
-		
-		goto out;
+		dbus_error_free (&error);
+		return;
 	}
 	
 	if (config.automount_media && config.autoplay_cda) {
-		action = gvm_prompt (GVM_PROMPT_CDA_EXTRA, 0, NULL);
+		ctx = gvm_prompt_ctx_new (GVM_PROMPT_CDA_EXTRA, cda_extra_cb, udi, device, NULL, NULL);
+		gvm_prompt (ctx, 0, NULL);
 	} else if (config.automount_media) {
-		action = GVM_RESPONSE_BROWSE;
-	} else if (config.autoplay_cda) {
-		action = GVM_RESPONSE_PLAY;
-	}
-	
-	switch (action) {
-	case GVM_RESPONSE_BROWSE:
 		if (!gvm_udi_is_subfs_mount (udi))
-			gvm_device_mount (udi, device, NULL);
-		break;
-	case GVM_RESPONSE_PLAY:
+			gvm_device_mount (udi, device, NULL, TRUE);
+	} else if (config.autoplay_cda) {
 		gvm_run_cdplayer (udi, device, NULL);
-		break;
-	default:
-		break;
 	}
 	
- out:
 	libhal_free_string (device);
-	libhal_free_string (mount_point);
 }
 
 enum {
@@ -1185,24 +1332,10 @@ gvm_device_is_writer (const char *udi)
 	return WRITER_TYPE_NONE;
 }
 
-
-/*
- * gvm_run_cdburner - execute the user-specified CD burner command on the
- * given device node, if so configured
- */
 static void
-gvm_run_cdburner (const char *udi, int type, const char *device, const char *mount_point)
+burn_cdr_cb (GvmPromptCtx *ctx, int action)
 {
 	const char *command;
-	int action;
-	
-	if (!config.autoburn)
-		return;
-	
-	if (type == WRITER_TYPE_DVD)
-		action = gvm_prompt (GVM_PROMPT_BURN_DVD, 0, NULL);
-	else
-		action = gvm_prompt (GVM_PROMPT_BURN_CDR, 0, NULL);
 	
 	switch (action) {
 	case GVM_RESPONSE_BURN_AUDIO_CD:
@@ -1215,7 +1348,29 @@ gvm_run_cdburner (const char *udi, int type, const char *device, const char *mou
 		return;
 	}
 	
-	gvm_run_command (command, udi, device, mount_point);
+	gvm_run_command (command, ctx->udi, ctx->device, ctx->mount_point);
+}
+
+/*
+ * gvm_run_cdburner - execute the user-specified CD burner command on the
+ * given device node, if so configured
+ */
+static void
+gvm_run_cdburner (const char *udi, int type, const char *device, const char *mount_point)
+{
+	GvmPromptCtx *ctx;
+	GvmPrompt prompt;
+	
+	if (!config.autoburn)
+		return;
+	
+	if (type == WRITER_TYPE_DVD)
+		prompt = GVM_PROMPT_BURN_DVD;
+	else
+		prompt = GVM_PROMPT_BURN_CDR;
+	
+	ctx = gvm_prompt_ctx_new (prompt, burn_cdr_cb, udi, device, mount_point, NULL);
+	gvm_prompt (ctx, 0, NULL);
 }
 
 
@@ -1243,8 +1398,7 @@ gvm_cdrom_policy (const char *udi)
 	
 	if (!device) {
 		warn ("cannot get block.device: %s", error.message);
-		if (dbus_error_is_set (&error))
-			dbus_error_free (&error);
+		dbus_error_free (&error);
 		goto out;
 	}
 	
@@ -1255,7 +1409,7 @@ gvm_cdrom_policy (const char *udi)
 		gvm_ask_mixed (udi);
 	} else if (has_data) {
 		if (config.automount_media && !gvm_udi_is_subfs_mount (udi))
-			gvm_device_mount (udi, device, NULL);
+			gvm_device_mount (udi, device, NULL, TRUE);
 	} else if (is_blank) {
 		if ((type = gvm_device_is_writer (drive_udi)))
 			gvm_run_cdburner (udi, type, device, NULL);
@@ -1264,6 +1418,7 @@ gvm_cdrom_policy (const char *udi)
 	/** @todo enforce policy for all the new disc types now supported */
 	
  out:
+	
 	libhal_free_string (device);
 	libhal_free_string (drive_udi);
 }
@@ -1302,8 +1457,7 @@ gvm_media_changed (const char *udi, const char *storage_device)
 	 */
 	if (!(media_type = libhal_device_get_property_string (hal_ctx, storage_device, "storage.drive_type", &error))) {
 		warn ("cannot get storage.drive_type: %s", error.message);
-		if (dbus_error_is_set (&error))
-			dbus_error_free (&error);
+		dbus_error_free (&error);
 		
 		/* FIXME: should we really return TRUE here? not sure... need a test case */
 		
@@ -1349,6 +1503,202 @@ gvm_automount_enabled (const char *udi)
 	return enabled;
 }
 
+static void
+import_ptp_camera_cb (GvmPromptCtx *ctx, int action)
+{
+	if (action == GVM_RESPONSE_IMPORT_PHOTOS)
+		gvm_run_camera (ctx->udi, NULL, NULL);
+}
+
+
+static int
+strptrcmp (const void *strptr0, const void *strptr1)
+{
+	return strcmp (*((const char **) strptr0), *((const char **) strptr1));
+}
+
+
+typedef gboolean (* DeviceAddedHandler) (const char *udi, const char *capability);
+
+static gboolean
+block_device_added (const char *udi, const char *capability __attribute__((__unused__)))
+{
+	char *fsusage = NULL, *device = NULL, *storage_device = NULL;
+	DBusError error;
+	int mountable;
+
+	dbus_error_init (&error);
+
+	/* is this a mountable volume? */
+	if (!(mountable = libhal_device_get_property_bool (hal_ctx, udi, "block.is_volume", NULL))) {
+		dbg ("not a mountable volume: %s\n", udi);
+		goto out;
+	}
+	
+	/* if it is a volume, it must have a device node */
+	if (!(device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error))) {
+		dbg ("cannot get block.device: %s\n", error.message);
+		goto out;
+	}
+	
+	if (mountable) {
+		/* only mount if the block device has a sensible filesystem */
+		fsusage = libhal_device_get_property_string (hal_ctx, udi, "volume.fsusage", &error);
+		if (!fsusage || strcmp (fsusage, "filesystem") != 0) {
+			dbg ("no sensible filesystem for %s\n", udi);
+			mountable = FALSE;
+		}
+	}
+	
+	/* get the backing storage device */
+	if (!(storage_device = libhal_device_get_property_string (hal_ctx, udi, "block.storage_device", &error))) {
+		dbg ("cannot get block.storage_device: %s\n", error.message);
+		goto out;
+	}
+	
+	/*
+	 * Does this device support removable media?  Note that we
+	 * check storage_device and not our own UDI
+	 */
+	if (libhal_device_get_property_bool (hal_ctx, storage_device, "storage.removable", NULL)) {
+		/* we handle media change events separately */
+		dbg ("Changed: %s\n", device);
+		if (gvm_media_changed (udi, storage_device))
+			goto out;
+	}
+	
+	if (config.automount_drives && mountable) {
+		if (!gvm_udi_is_subfs_mount (udi)) {
+			if (gvm_automount_enabled (udi)) {
+				gvm_device_mount (udi, device, NULL, TRUE);
+			} else {
+				dbg ("storage.automount_enabled_hint set to false on %s, not mounting\n", udi);
+			}
+		}
+	}
+	
+out:
+	
+	if (dbus_error_is_set (&error))
+		dbus_error_free (&error);
+	
+	libhal_free_string (device);
+	libhal_free_string (fsusage);
+	libhal_free_string (storage_device);
+	
+	return TRUE;
+}
+
+static gboolean
+camera_device_added (const char *udi, const char *capability __attribute__((__unused__)))
+{
+	GvmPromptCtx *ctx;
+	
+	if (!gvm_udi_is_ptp_camera (udi))
+		return FALSE;
+	
+	if (!(config.autophoto && config.autophoto_command))
+		return TRUE;
+	
+	ctx = gvm_prompt_ctx_new (GVM_PROMPT_IMPORT_PTP_CAMERA, import_ptp_camera_cb, udi, NULL, NULL, NULL);
+	gvm_prompt (ctx, 0, NULL);
+	
+	return TRUE;
+}
+
+static struct {
+	const char *capability;
+	gboolean *autoexec;
+	char **command;
+} inputs[] = {
+	{ "input.keyboard", &config.autokeyboard, &config.autokeyboard_command },
+	{ "input.mouse",    &config.automouse,    &config.automouse_command    },
+	{ "input.tablet",   &config.autotablet,   &config.autotablet_command   },
+};
+
+static gboolean
+input_device_added (const char *udi, const char *capability)
+{
+	/* input device (keyboard, mouse, wacom tablet, etc...) */
+	const char *command = NULL;
+	int autoexec = FALSE;
+	int handled = FALSE;
+	DBusError error;
+	char *device;
+	size_t i;
+	
+	for (i = 0; i < G_N_ELEMENTS (inputs); i++) {
+		if (!strcmp (inputs[i].capability, capability)) {
+			autoexec = *inputs[i].autoexec;
+			command = *inputs[i].command;
+			handled = TRUE;
+			break;
+		}
+	}
+	
+	if (autoexec && command) {
+		dbus_error_init (&error);
+		if ((device = libhal_device_get_property_string (hal_ctx, udi, "input.device", &error))) {
+			gvm_run_command (command, udi, device, NULL);
+			libhal_free_string (device);
+		} else {
+			warn ("cannot get input.device property: %s", error.message);
+			dbus_error_free (&error);
+		}
+	}
+	
+	return handled;
+}
+
+static gboolean
+pda_device_added (const char *udi, const char *capability __attribute__((__unused__)))
+{
+	DBusError error;
+	char *platform;
+	
+	dbus_error_init (&error);
+	
+	if ((platform = libhal_device_get_property_string (hal_ctx, udi, "pda.platform", &error))) {
+		if (!strcmp (platform, "palm")) {
+			if (config.autopilot)
+				gvm_run_pilot (udi);
+		} else if (!strcmp (platform, "pocketpc")) {
+			if (config.autopocketpc)
+				gvm_run_pocketpc (udi);
+		}
+		
+		libhal_free_string (platform);
+	} else {
+		warn ("cannot get pda.platform property: %s", error.message);
+		dbus_error_free (&error);
+	}
+	
+	return TRUE;
+}
+
+static gboolean
+printer_device_added (const char *udi, const char *capability __attribute__((__unused__)))
+{
+	if (config.autoprinter)
+		gvm_run_printer (udi);
+	
+	return TRUE;
+}
+
+static struct {
+	const char *capability;
+	DeviceAddedHandler handler;
+} devices[] = {
+	{ "block",          block_device_added   },
+	{ "camera",         camera_device_added  },
+	/*{ "input",          input_device_added   },*/ /* we don't handle generic input devices (yet?) */
+	{ "input.keyboard", input_device_added   },
+	{ "input.mouse",    input_device_added   },
+	{ "input.tablet",   input_device_added   },
+	{ "pda",            pda_device_added     },
+	{ "printer",        printer_device_added },
+};
+
 /** Invoked when a device is added to the Global Device List. 
  *
  *  @param  ctx                 LibHal context
@@ -1358,101 +1708,33 @@ static void
 hal_device_added (LibHalContext *ctx __attribute__((__unused__)),
 		  const char *udi)
 {
-	char *fsusage = NULL, *device = NULL, *storage_device = NULL;
-	DBusError error;
-	int mountable;
+	char **capabilities;
+	size_t i, j, n;
 	
-	dbg ("New Device: %s\n", udi);
+	dbg ("Device added: %s\n", udi);
 	
-	dbus_error_init (&error);
-	if (libhal_device_query_capability (hal_ctx, udi, "block", NULL)) {
-		/* is this a mountable volume? */
-		if (!(mountable = libhal_device_get_property_bool (hal_ctx, udi, "block.is_volume", NULL))) {
-			dbg ("not a mountable volume: %s\n", udi);
-		}
+	if (!(capabilities = libhal_device_get_property_strlist (hal_ctx, udi, "info.capabilities", NULL)))
+		return;
+	
+	for (n = 0; capabilities[n]; n++)
+		;
+	
+	qsort (capabilities, n, sizeof (char *), strptrcmp);
+	
+	for (i = 0, j = 0; i < G_N_ELEMENTS (devices) && j < n; i++) {
+		int cmp = -1;
+
+		while (j < n && (cmp = strcmp (capabilities[j], devices[i].capability)) < 0)
+			j++;
 		
-		/* if it is a volume, it must have a device node */
-		if (!(device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error))) {
-			dbg ("cannot get block.device: %s\n", error.message);
-			goto out;
+		if (cmp == 0) {
+			if (devices[i].handler (udi, capabilities[j]))
+				break;
+			j++;
 		}
-		
-		if (mountable) {
-			/* only mount if the block device has a sensible filesystem */
-			fsusage = libhal_device_get_property_string (ctx, udi, "volume.fsusage", &error);
-			if (!fsusage || strcmp (fsusage, "filesystem") != 0) {
-				dbg ("no sensible filesystem for %s\n", udi);
-				mountable = FALSE;
-			}
-		}
-		
-		/* get the backing storage device */
-		if (!(storage_device = libhal_device_get_property_string (hal_ctx, udi, "block.storage_device", &error))) {
-			dbg ("cannot get block.storage_device: %s\n", error.message);
-			goto out;
-		}
-		
-		/*
-		 * Does this device support removable media?  Note that we
-		 * check storage_device and not our own UDI
-		 */
-		if (libhal_device_get_property_bool (hal_ctx, storage_device, "storage.removable", NULL)) {
-			/* we handle media change events separately */
-			dbg ("Changed: %s\n", device);
-			if (gvm_media_changed (udi, storage_device))
-				goto out;
-		}
-		
-		if (config.automount_drives && mountable) {
-			if (!gvm_udi_is_subfs_mount (udi)) {
-				if (gvm_automount_enabled (udi)) {
-					gvm_device_mount (udi, device, NULL);
-				} else {
-					dbg ("storage.automount_enabled_hint set to false on %s, not mounting\n", udi);
-				}
-			}
-		}
-	} else if (libhal_device_query_capability (hal_ctx, udi, "input", NULL)) {
-		/* input device (keyboard, mouse, wacom tablet, etc...) */
-		const char *command = NULL;
-		int autoexec = FALSE;
-		char *device;
-		
-		if (libhal_device_query_capability (hal_ctx, udi, "input.keyboard", NULL)) {
-			autoexec = config.autokeyboard;
-			command = config.autokeyboard_command;
-		} else if (libhal_device_query_capability (hal_ctx, udi, "input.mouse", NULL)) {
-			autoexec = config.automouse;
-			command = config.automouse_command;
-		} else if (libhal_device_query_capability (hal_ctx, udi, "input.tablet", NULL)) {
-			autoexec = config.autotablet;
-			command = config.autotablet_command;
-		}
-		
-		if (autoexec && command) {
-			if ((device = libhal_device_get_property_string (hal_ctx, udi, "input.device", &error))) {
-				gvm_run_command (command, udi, device, NULL);
-				libhal_free_string (device);
-			} else {
-				warn ("cannot get input.device property: %s", error.message);
-			}
-		}
-	} else if (libhal_device_query_capability (hal_ctx, udi, "printer", NULL)) {
-		if (config.autoprinter)
-			gvm_run_printer (udi);
-	} else if (gvm_udi_is_ptp_camera (udi)) {
-		/* if the device is a PTP camera with libgphoto2 support, launch the user-specified application */
-		if (config.autophoto && gvm_prompt (GVM_PROMPT_IMPORT_PTP_CAMERA, 0, NULL) == GVM_RESPONSE_IMPORT_PHOTOS)
-			gvm_run_camera (udi, NULL, NULL);
 	}
 	
- out:
-	if (dbus_error_is_set (&error))
-		dbus_error_free (&error);
-	
-	libhal_free_string (device);
-	libhal_free_string (fsusage);
-	libhal_free_string (storage_device);
+	libhal_free_string_array (capabilities);
 }
 
 /** Invoked when a device is removed from the Global Device List. 
@@ -1507,33 +1789,37 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 		       dbus_bool_t is_removed __attribute__((__unused__)), 
 		       dbus_bool_t is_added __attribute__((__unused__)))
 {
-	char *mounting_udi;
-	dbus_bool_t val;
+	struct _MountPolicy *policy;
+	gboolean mounted;
 	GSList *l, *n;
 	
 	if (strcmp (key, "volume.is_mounted") != 0)
 		return;
 	
-	val = libhal_device_get_property_bool (hal_ctx, udi, key, NULL);
+	mounted = libhal_device_get_property_bool (hal_ctx, udi, key, NULL);
 	
-	if (val == TRUE) {
+	if (mounted) {
 		dbg ("Mounted: %s\n", udi);
 		
-		if ((mounting_udi = g_hash_table_lookup (mount_table, udi))) {
+		if ((policy = g_hash_table_lookup (mount_table, udi))) {
 			g_hash_table_remove (mount_table, udi);
-			g_free (mounting_udi);
 			
 			/* add to list of all volumes mounted during lifetime */
 			mounted_volumes = g_slist_append (mounted_volumes, g_strdup (udi));
+			
+			if (policy->apply)
+				gvm_device_mounted (udi);
+			
+			g_free (policy->udi);
+			g_free (policy);
 		} else {
 			dbg ("not in mount queue: %s\n", udi);
+			gvm_device_mounted (udi);
 		}
-		
-		gvm_device_mounted (udi);
 	} else {
 		dbg ("Unmounted: %s\n", udi);
 		
-		/* remove from list of all volumes mounted during lifetime */
+		/* unmount all volumes mounted during lifetime */
 		
 		for (l = mounted_volumes; l != NULL; l = n) {
 			n = g_slist_next (l);
@@ -1717,7 +2003,7 @@ mount_all (LibHalContext *ctx)
 		/* mount the device */
 		if ((dev = libhal_device_get_property_string (ctx, udi, "block.device", &error))) {
 			dbg ("mount_all: mounting %s\n", dev);
-			gvm_device_mount (udi, dev, NULL);
+			gvm_device_mount (udi, dev, NULL, FALSE);
 			libhal_free_string (dev);
 		} else {
 			warn ("mount_all: no device for udi=%s: %s", udi, error.message);
@@ -1813,6 +2099,25 @@ gvm_die (GnomeClient *client __attribute__((__unused__)),
 	gtk_main_quit ();
 }
 
+/* assume that display :0 is the system console */
+static int
+gvm_running_on_console (void)
+{
+	const char *str;
+	
+	if (!(str = DisplayString (GDK_DISPLAY ())))
+		return TRUE;
+	
+	if (!(str = strchr (str, ':')))
+		return TRUE;
+	
+	str++;
+	if (str[0] == ':')
+		str++;
+	
+	return strtoul (str, NULL, 10) == 0;
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -1826,18 +2131,22 @@ main (int argc, char *argv[])
 	textdomain (PACKAGE);
 	
 	client = gnome_master_client ();
-	if (gvm_get_clipboard ()) {
+	if (gvm_get_clipboard () && gvm_running_on_console ()) {
 		gnome_client_set_restart_style (client, GNOME_RESTART_ANYWAY);
 	} else {
 		gnome_client_set_restart_style (client, GNOME_RESTART_NEVER);
-		warn ("already running?");
+		
+		if (!gvm_running_on_console ())
+			warn ("not on the system console");
+		else
+			warn ("already running");
+		
 		return 1;
 	}
 	
 	g_signal_connect (client, "die", G_CALLBACK (gvm_die), NULL);
 	
-	hal_ctx = gvm_do_hal_init ();
-	if (!hal_ctx)
+	if (!(hal_ctx = gvm_do_hal_init ()))
 		return 1;
 	
 	gvm_init_config ();
